"use strict";

Object.defineProperty(exports, "__esModule", { value: true });
var Immutable = require("immutable");
var redux_1 = require("redux");
var dob_1 = require("dob");
var utils_1 = require("./utils");
/**
 * observable 包裹的对象，对应 immutable 结构
 */
var immutables = new Map();
/**
 * 每一个字对象，存储其根节点以及路径
 */
var nodeInfos = new Map();
/**
 * 存储所有 observer 对象对应的 callback
 */
var snapshots = new Map();
/**
 * 标记为 async method
 */
var asyncSymbol = Symbol();
/**
 * 初始化 immutable 对象
 */
function initImmutable(obj) {
    if (immutables.has(obj)) {
        return;
    }
    switch (obj.constructor) {
        case Array:
            immutables.set(obj, Immutable.fromJS(obj));
            break;
        case Object:
            immutables.set(obj, Immutable.fromJS(obj));
            break;
        default:
            // 剩下自定义 class new 出来的，需要特殊处理才能解析
            immutables.set(obj, Immutable.fromJS(JSON.parse(JSON.stringify(obj))));
    }
}
exports.initImmutable = initImmutable;
/**
 * 为对象子属性记录其根对象、访问路径
 * 当访问 setter 后，我们可以找到其根对象和路径，采取 immutable 更新
 */
dob_1.globalState.event.on("get", function (info) {
    var obj = info.target;
    // 先认为根对象就是自己
    var rootObj = obj;
    Object.keys(obj).forEach(function (key) {
        // 当前对象的路径
        var paths = [];
        // 先判断自身是否有父元素
        if (nodeInfos.has(obj)) {
            var nodeInfo = nodeInfos.get(obj);
            rootObj = nodeInfo.rootObject;
            paths = nodeInfo.paths.slice();
        }
        var child = obj[key];
        // 不处理基本类型
        if (utils_1.isPrimitive(child)) {
            return;
        }
        paths.push(key);
        nodeInfos.set(child, {
            rootObject: rootObj,
            paths: paths
        });
    });
});
/**
 * 注销某个 object 的 nodeInfos 数据
 */
function removeNodeInfo(obj) {
    if (utils_1.isPrimitive(obj)) {
        return;
    }
    nodeInfos.delete(obj);
}
/**
 * 当任何对象触发 set 时
 */
dob_1.globalState.event.on("set", function (info) {
    var rootObj = info.target;
    var paths = [info.key.toString()];
    // 如果找到了，说明修改的是叶子元素
    if (nodeInfos.has(info.target)) {
        var nodeInfo = nodeInfos.get(info.target);
        rootObj = nodeInfo.rootObject;
        paths = nodeInfo.paths.concat(info.key.toString());
    }
    if (!immutables.has(rootObj)) {
        // 找不到根对象，说明很可能是没有使用 createReduxStore 注册，那就不管了
        return;
    }
    var rootImmutableObj = immutables.get(rootObj);
    // TODO: 删除原值 value 的所有子属性，包括自己的 nodeInfos
    // 生成新的 immutable 对象
    var newImmutableObj = rootImmutableObj.setIn(paths, Immutable.fromJS(info.value));
    // 更新 immutable 对象
    immutables.set(rootObj, newImmutableObj);
    // 找到这个对象的 snapshotCallback 并触发
    if (dob_1.globalState.proxies.has(rootObj)) {
        var proxy = dob_1.globalState.proxies.get(rootObj);
        if (snapshots.has(proxy)) {
            var snapshot = snapshots.get(proxy);
            var objectJs = newImmutableObj.toJS();
            snapshot.forEach(function (each) {
                each(objectJs);
            });
        }
    }
});
/**
 * 当触发任何对象 delete 时
 */
dob_1.globalState.event.on("deleteProperty", function (info) {
    var rootObj = info.target;
    var paths = [info.key.toString()];
    // 如果找到了，说明修改的是叶子元素
    if (nodeInfos.has(info.target)) {
        var nodeInfo = nodeInfos.get(info.target);
        rootObj = nodeInfo.rootObject;
        paths = nodeInfo.paths.concat(info.key.toString());
    }
    if (!immutables.has(rootObj)) {
        // 找不到根对象，说明很可能是没有使用 createReduxStore 注册，那就不管了
        return;
    }
    var rootImmutableObj = immutables.get(rootObj);
    // TODO: 删除原值 value 的所有子属性，包括自己的 nodeInfos
    // 生成新的 immutable 对象
    var newImmutableObj = rootImmutableObj.deleteIn(paths);
    // 更新 immutable 对象
    immutables.set(rootObj, newImmutableObj);
    // 找到这个对象的 snapshotCallback 并触发
    if (dob_1.globalState.proxies.has(rootObj)) {
        var proxy = dob_1.globalState.proxies.get(rootObj);
        if (snapshots.has(proxy)) {
            var snapshot = snapshots.get(proxy);
            var objectJs = newImmutableObj.toJS();
            snapshot.forEach(function (each) {
                each(objectJs);
            });
        }
    }
});
/**
 * 监听对象快照
 */
function onSnapshot(proxy, callback) {
    var obj = dob_1.globalState.originObjects.get(proxy);
    if (!obj) {
        throw Error("请传入 observable 函数返回的对象");
    }
    // 初始化 immutable，从此只要这个对象变动，就会生成新 immutable
    initImmutable(obj);
    if (!snapshots.has(proxy)) {
        snapshots.set(proxy, new Set());
    }
    var snapshot = snapshots.get(proxy);
    snapshot.add(callback);
}
exports.onSnapshot = onSnapshot;
/**
 * 获取对象的快照
 */
function getSnapshot(proxyObj) {
    var obj = dob_1.globalState.originObjects.get(proxyObj);
    return immutables.get(obj).toJS();
}
exports.getSnapshot = getSnapshot;
/**
 * 创建 redux store
 */
function createReduxStore(stores, enhancer) {
    var combineActions = {};
    var reducers = Object.keys(stores).reduce(function (allReducers, key) {
        var storeClass = stores[key];
        var storeInstance = new storeClass(); // 修正定义
        // 用来 dispatch 的 actions
        var actions = {};
        // class 中的 store
        var observableStore = null;
        // 直接遍历获取 store
        for (var field in storeInstance) {
            if (storeInstance.hasOwnProperty(field)) {
                var value = storeInstance[field];
                // 如果是 observable，说明这是个 store
                if (dob_1.isObservable(value)) {
                    observableStore = value;
                    var obj = dob_1.globalState.originObjects.get(value);
                    // 初始化 immutable，从此只要这个对象变动，就会生成新 immutable
                    initImmutable(obj);
                }
            }
        }
        // 遍历 proto 获取所有 action
        // 将 action 执行的内容放在对应 reducer 中，并且 action 只 dispatch 不执行 observer
        Object.getOwnPropertyNames(Object.getPrototypeOf(storeInstance)).filter(function (methodName) {
            return methodName !== "constructor";
        }).forEach(function (methodName) {
            if (storeInstance[asyncSymbol] && storeInstance[asyncSymbol].has(methodName)) {
                // 对于异步函数，是有副作用的，这里不能修改 store，修改了会报错
                // 直接执行没有问题，需要 return
                actions[methodName] = storeInstance[methodName];
            } else {
                // 新增 actions，用来 dispatch
                actions[methodName] = function () {
                    for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
                        args[_key] = arguments[_key];
                    }

                    return {
                        type: key + "." + methodName,
                        payload: args
                    };
                };
            }
        });
        combineActions[key] = actions;
        // ... 用法是为了取消引用问题，因为直接修改原对象，初始状态应该是个快照
        allReducers[key] = function () {
            var state = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : Object.assign({}, observableStore);
            var action = arguments[1];

            if (action.type.indexOf(".") === -1) {
                return state;
            }
            // 从 type 剥离出 methodName
            var types = action.type.split(".");
            types.shift();
            var methodName = types.join(".");
            // 如果不在 actions 中，就忽略
            if (!Object.keys(actions).find(function (actionMethodName) {
                return methodName === actionMethodName;
            })) {
                return state;
            }
            // 执行 action 中当前方法，此时会生成快照
            storeInstance[methodName].apply(storeInstance, action.payload);
            // 返回快照
            return getSnapshot(observableStore);
        };
        return allReducers;
    }, {});
    var rootReducer = redux_1.combineReducers(reducers);
    var composeEnhancers = window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__ || redux_1.compose;
    var store = enhancer ? redux_1.createStore(rootReducer, composeEnhancers(redux_1.applyMiddleware(enhancer, utils_1.createThunkMiddleware))) : redux_1.createStore(rootReducer, composeEnhancers(redux_1.applyMiddleware(utils_1.createThunkMiddleware)));
    return {
        store: store,
        /**
         * action 集合，已经绑定了 dispatch，直接调用
         */
        actions: Object.keys(combineActions).reduce(function (actions, key) {
            var waitBindAction = combineActions[key];
            actions[key] = redux_1.bindActionCreators(waitBindAction, store.dispatch);
            return actions;
        }, {})
    };
}
exports.createReduxStore = createReduxStore;
/**
 * 标记 method 为 async
 */
exports.Task = function (target, propertyKey, descriptor) {
    if (!target[asyncSymbol]) {
        target[asyncSymbol] = new Set();
    }
    target[asyncSymbol].add(propertyKey);
};
//# sourceMappingURL=immutable.js.map